﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using ProcessRubyScript;
using System.Security.Cryptography;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using CodeRetrieverDataflow;

namespace DataFlowConverter
{
    public enum blockType { zeroHash, retrieved, notRetrieved };

    public class ConvertRuby
    {
        public ConvertRuby(ref XDocument xmlXDoc, bool genVars)
        {
            _varFlag = genVars;
            _varList = new List<string>();
            _profileList = new List<string>();
            _truncList = new List<string>();
            _rubyExprList = new List<string>();
            _rubyHashList = new List<string>();
            _notRetBlkList = new List<Block>();
            _retBlkList = new List<Block>();
            _zeroBlkList = new List<Block>();
            _rubyExp = new RubyExpression();

            xmlDoc = xmlXDoc;

            convertStr = null;
            conversionError = false;

            // Debug Only:
            _filterList = new List<string>();
        }

        public void DoConversion()
        {
            InitLists();

            _rubyExprList = _rubyExp.GetRubyExpressions();

            if ((_rubyExprList != null) && (_rubyExprList.Count > 0))
            {
                /*
                 * Now compute a hash for each Ruby expression, then
                 * retrieve the JavaScript code based on mnemonic hashes
                 * which are constructed from the expression hashes. Lots of hash.
                */
                if (ComputeExprHashes())
                {
                    if ((_rubyHashList != null) && (_rubyHashList.Count > 0))
                    {
                        if (_rubyExprList.Count == _rubyHashList.Count)
                        {
                            /*
                             * This will be null when initially creating the
                             * js version of the Ruby snippet file so we would
                             * stop here.
                            */
                            if (xmlDoc != null)
                            {
                                if (SetBlocks())
                                {
                                    SetConvertStr();
                                }
                            }
                        }
                        else
                        {
                            MessageBox.Show("_rubyExprList.Count != _rubyHashList.Count.", "DoConversion() list size differences:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            return;
                        }
                    }
                }

                // DEBUG:
                //MessageBox.Show(String.Format("_rubyExprList.Count = {0}\n\n_rubyHashList.Count = {1}\n\n_filterList.Count = {2}\n", _rubyExprList.Count, _rubyHashList.Count, _filterList.Count));
            }
        }

        public void SetConvertStr()
        {
            /*
             * The final conversion step: Set the class variable convertStr, write it to the 
             * clipboard. If we have any profile declarations we output those first, followed
             * by any var declarations if flagged for it. We also check for not retrieved blocks
             * and set the class variable conversionError to true if we encounter any.
            */
            bool process = true;
            bool processNotRet;
            bool processRet;
            bool processZero;

            int notRetBlkListIndex = 0;
            int retBlkListIndex = 0;
            int zeroBlkListIndex = 0;

            int notRetLast;
            int retLast;
            int zeroLast;
            int retFirst;
            int smallestIndex;

            string smallestList = "";
            string profileVarStr = "";

            try
            {
                while (process)
                {
                    notRetLast = -1;
                    retFirst = -1;
                    retLast = -1;
                    zeroLast = -1;
                    smallestIndex = -1;
                    smallestList = "";

                    processNotRet = (notRetBlkListIndex < _notRetBlkList.Count);
                    processRet = (retBlkListIndex < _retBlkList.Count);
                    processZero = (zeroBlkListIndex < _zeroBlkList.Count);

                    if (processNotRet)
                    {
                        notRetLast = _notRetBlkList[notRetBlkListIndex].hashIndices.Last();
                    }

                    if (processRet)
                    {
                        retFirst = _retBlkList[retBlkListIndex].hashIndices.First();
                        retLast = _retBlkList[retBlkListIndex].hashIndices.Last();
                    }

                    if (processZero)
                    {
                        zeroLast = _zeroBlkList[zeroBlkListIndex].hashIndices.Last();
                    }

                    // Find smallest non-negative index value:
                    if (notRetLast > -1)
                    {
                        smallestIndex = notRetLast;
                        smallestList = "notRet";
                    }

                    if (retLast > -1)
                    {
                        if ((smallestIndex < 0) || (retLast < smallestIndex))
                        {
                            smallestIndex = retLast;
                            smallestList = "ret";
                        }
                    }

                    if (zeroLast > -1)
                    {
                        if ((smallestIndex < 0) || (zeroLast < smallestIndex))
                        {
                            // New approach, avoiding blank lines that lie within a Ruby expression --- start.
                            if ((smallestList == "ret") && (zeroLast < smallestIndex) && (zeroLast > retFirst) && (_rubyExprList[zeroLast].Trim().Length == 0) && (_zeroBlkList[zeroBlkListIndex].hashIndices.Count == 1))
                            {
                                /*
                                 * 'Consume' this index, as it actually is a single blank line inside a Ruby
                                 * expression. We need to do this else we would be inserting a blank line
                                 * ahead of the retrieved block's JavaScript code.
                                */
                                zeroBlkListIndex += 1;
                            }
                            else
                            {
                                smallestIndex = zeroLast;
                                smallestList = "zero";
                            }
                            // New approach, avoiding blank lines that lie within a Ruby expression --- stop.

                            // I was just doing this without the above new approach.
                            //smallestIndex = zeroLast;
                            //smallestList = "zero";
                        }
                    }

                    /*
                     * Now concatenate to convertStr the block's outputStr from the
                     * list that has the lowest last hash index value for it's associated
                     * current index value.
                    */
                    switch (smallestList)
                    {
                        case ("notRet"):

                            convertStr += _notRetBlkList[notRetBlkListIndex].outputStr;
                            notRetBlkListIndex += 1;
                            break;

                        case ("ret"):

                            convertStr += _retBlkList[retBlkListIndex].outputStr;
                            retBlkListIndex += 1;
                            break;

                        case ("zero"):

                            convertStr += _zeroBlkList[zeroBlkListIndex].outputStr;
                            zeroBlkListIndex += 1;
                            break;

                        default:

                            process = false;
                            break;
                    }
                }

                // Profile declarations:
                if (_profileList.Count > 0)
                {
                    profileVarStr = string.Join("", _profileList);
                }

                // Variable declarations:
                if (_varList.Count > 0)
                {
                    profileVarStr = ((profileVarStr.Length > 0 ? profileVarStr + Environment.NewLine : "") + string.Join("", _varList));
                }

                if (profileVarStr.Length > 0)
                {
                    if (!String.IsNullOrEmpty(convertStr))
                    {
                        profileVarStr += (convertStr.StartsWith(Environment.NewLine) ? "" : Environment.NewLine);
                    }
                }

                // Now put the profileVarStr first:
                convertStr = profileVarStr + convertStr;

                if (String.IsNullOrEmpty(convertStr))
                {
                    Clipboard.Clear();
                }
                else
                {
                    /*
                     * Removing any extra blank lines that may have creeped in, for example
                     * that would be tha case when processing profile declarations when there is
                     * a blank line both above and below a profile instantiation.
                    */
                    convertStr = Regex.Replace(convertStr, @"(\r\n){3,}", "\r\n\r\n");

                    // Clipboard.SetText first clears the Clipboard.
                    Clipboard.SetText(convertStr, TextDataFormat.Text);

                    conversionError = _notRetBlkList.Count > 0;

                    // Just a warning, most likely we will not be at 100% conversion.
                    if (conversionError)
                    {
                        MessageBox.Show("Some Ruby code was not converted.\n\nThis code has been commented out.\n\nIt may still be possible to obtain the JavaScript equivalent code by list selection. This would be the case when the Ruby code is an older version not used by the Retriever. You will need to switch to 'Retrieve JavaScript' mode before you try to retrieve the newer JavaScript version of the Ruby code.\n\nPlease email Harry with the Ruby code that failed to convert along with the doc id.\n\nALL conversion results, pass or fail, have also been copied to the clipboard.", "Conversion Failure(s):", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                Clipboard.Clear();
                convertStr = null;
            }
        }

        /*
         * Sets the private class variables:
         * 
         * _notRetBlkList - 'Not Retrieved Block List':
         *      A block is set for each Ruby expression that has not been converted.
         *      Contiguous expressions [indices] are blocked together in one block.
         *      
         * _retBlkList - 'Retrieved Block List':
         *      A block is set for each Ruby expression corresponding to a retrievable 
         *      mnemonic. Javascript is retrieved as well as the boxes, which are renumbered
         *      per the Ruby code, truncation and line indentation are also restored if present
         *      in the Ruby code. We also add any variables to the class member _varList.
         *      
         * _zeroHashBlkList - A block is set for blank and comment line Ruby expressions. 
         *      These have a hash value of "0". Contiguous expressions [indices] are blocked
         *      together in one block.
        */
        //public bool SetBlocks()
        //{
        //    if (ConvertRuby_To_JS.Form1.debug)
        //    {
        //        return false;
        //    }

        //    /*
        //     * Caller has ensured that _rubyExprList and _rubyHashList are not null, and 
        //     * contain at least 1 expression, and also that _rubyExprList.Count is equal to
        //     * _rubyHashList.Count.
        //    */
        //    Block block;
        //    Block retBlock;
        //    Block notRetBlock;

        //    string hashStr;
        //    string concatHashStr;
        //    bool processHashes;
        //    bool foundHash = false;
        //    int startIndex;
        //    int stopIndex;
        //    int lastNotRetIndex;
        //    int lastIndex;
        //    List<int> nonZeroIndices = new List<int>();
        //    List<int> notRetIndices = new List<int>();
        //    Regex commentRegex = new Regex("#");

        //    try
        //    {
        //        _notRetBlkList.Clear();
        //        _retBlkList.Clear();
        //        _zeroBlkList.Clear();

        //        /*
        //         * We will first set the list of non-zero hash indices of _rubyHashList.
        //         * These are the indices we will use in attempting to retrieve JavaScript for
        //         * the corresponding Ruby expressions in _rubyExprList. We also set the
        //         * zero hash block list [Ruby comment lines and blank lines] while we are here.
        //        */
        //        block = null;

        //        for (int i = 0; i < _rubyHashList.Count; i++)
        //        {
        //            if (_rubyHashList[i] == "0")
        //            {
        //                if (block == null)
        //                {
        //                    block = new Block(blockType.zeroHash);
        //                }

        //                block.hashIndices.Add(i);

        //                /*
        //                 * Expression is either a comment line in which case we convert the
        //                 * leading '#' to '//' or it is a blank line.
        //                */
        //                block.outputStr += commentRegex.Replace(_rubyExprList[i], "//", 1);
        //            }
        //            else
        //            {
        //                nonZeroIndices.Add(i);

        //                if (block != null)
        //                {
        //                    // _rubyHashList indices need to be contiguous for a zero block.
        //                    _zeroBlkList.Add(block);
        //                    block = null;
        //                }
        //            }
        //        }

        //        // Handle any hanging zeroBlock:
        //        if (block != null)
        //        {
        //            // _rubyHashList indices need to be contiguous for a zero block.
        //            _zeroBlkList.Add(block);
        //        }

        //        /*
        //         * We now set _notRetBlkList and _retBlkList by iterating over nonZeroIndices
        //         * until all indices are either in one block list or the other.
        //        */
        //        startIndex = 0;
        //        stopIndex = 0;
        //        lastNotRetIndex = 0;
        //        lastIndex = nonZeroIndices.Count - 1;

        //        retBlock = null;
        //        notRetBlock = null;

        //        processHashes = true;

        //        while (processHashes)
        //        {
        //            concatHashStr = "";
        //            hashStr = "";
        //            foundHash = false;

        //            for (int i = startIndex; i < nonZeroIndices.Count; i++)
        //            {
        //                concatHashStr += _rubyHashList[nonZeroIndices[i]];

        //                if (i > startIndex)
        //                {
        //                    hashStr = GetMnemonicHash(concatHashStr);
        //                }
        //                else
        //                {
        //                    hashStr = concatHashStr;
        //                }

        //                GetJSDataflowByHash(hashStr, ref retBlock);

        //                /*
        //                 * null block means we need to keep checking.
        //                */
        //                if (retBlock != null)
        //                {
        //                    /*
        //                     * At this point the block contains the Javascript code for the
        //                     * associated Ruby expression(s) as well as the generic boxes for
        //                     * the code snippet. The type has also been set to 'retrieved'.
        //                    */
        //                    stopIndex = i;
        //                    foundHash = true;
        //                    break;
        //                }
        //            }

        //            if (foundHash)
        //            {
        //                /*
        //                 * Note that some of these indices may not be contiguous.
        //                 * Reason being that somebody may have inserted say a comment line
        //                 * in between two expressions that comprise a code snippet.
        //                 * 
        //                 * So if we have indices 10,11, and 15 here we would expect to have indices
        //                 * 12, 13, and 14 in a block (1 block since they are contiguous) in the 
        //                 * _zeroBlkList which we would ouput FIRST, then output this converted block.
        //                 * We would output the _zeroBlkList first as we first check the last index
        //                 * value in the block's hashIndices list, refer to SetConvertStr().
        //                 * 
        //                 * Indices 12, 13, and 14 would NOT be blocked in the _notRetBlkList because
        //                 * these indices would then belong to non-zero hash expressions which would
        //                 * have 'broken' this conversion [since they were not found] - so the
        //                 * mnemonic hash of (index10Hash + index11Hash + index12Hash + index13Hash +
        //                 * index14Hash + index15Hash) could not possibly be found in the snippet file.
        //                */
        //                for (int j = startIndex; j <= stopIndex; j++)
        //                {
        //                    retBlock.hashIndices.Add(nonZeroIndices[j]);
        //                }

        //                /*
        //                 * We now need to: Set the generic boxes to their corresponding Ruby
        //                 * equivalents; Restore any truncation; Add any line indentation that
        //                 * is present in the Ruby code lines; Update a unique class variable name
        //                 * list; Insert this block into _retBlkList.
        //                */
        //                ProcessRubyRetBlock(ref retBlock);

        //                if (retBlock == null)
        //                {
        //                    // User has seen the exception message box so we just return.
        //                    return false;
        //                }

        //                startIndex = (stopIndex + 1);
        //            }
        //            else
        //            {
        //                /*
        //                 * We now know there is no correspnding Javascript code for the associated
        //                 * Ruby expression _rubyExprList[startIndex].
        //                */
        //                if (notRetBlock == null)
        //                {
        //                    notRetBlock = new Block(blockType.notRetrieved);
        //                }
        //                else if ((nonZeroIndices[startIndex] - lastNotRetIndex) > 1)
        //                {
        //                    /*
        //                     * We want contiguous indices for not retrieved blocks, we group comment
        //                     * contiguous Ruby expressions - the user sees the Ruby code but commented
        //                     * out with a conversion failure header.
        //                    */
        //                    ProcessNotRetBlock(ref notRetBlock);

        //                    notRetBlock = new Block(blockType.notRetrieved);
        //                }

        //                notRetBlock.hashIndices.Add(nonZeroIndices[startIndex]);
        //                lastNotRetIndex = nonZeroIndices[startIndex];

        //                startIndex += 1;
        //            }

        //            processHashes = (startIndex < nonZeroIndices.Count);
        //        }

        //        // Handle any hanging notRetBlock:
        //        if (notRetBlock != null)
        //        {
        //            ProcessNotRetBlock(ref notRetBlock);
        //        }

        //        return true;
        //    }
        //    catch (Exception ex)
        //    {
        //        MessageBox.Show(ex.ToString());
        //        return false;
        //    }
        //}

        // Trying a non-contiguous notRetBlock approach. ALSO HAD TO CHANGE ProcessNotRetBlock().
        /*
         * Sets the private class variables:
         * 
         * _notRetBlkList - 'Not Retrieved Block List':
         *      A block is set for each Ruby expression that has not been converted.
         *      However non-contiguous expressions [indices] will be blocked together
         *      in one block. This approach prevents a conversion failure indication
         *      for 2 Ruby expressions that are separated by a blank line but are
         *      actually 1 distinct code snippet - Sole Proprietor address dataflow
         *      would be a good example, blank lines are built-in to the code snippet.
         *      If the snippet described above failed to convert, the old approach, which
         *      was based on contiguous indices, would have a conversion failure message
         *      for indices 0 and 1, and another conversion failure message for indices
         *      3 and 4, because both sets have contiguous numbering. I did not like that, so
         *      we are allowing noncontiguous index handling for snippets that fail to convert.
         *      
         *      The old way would handle a conversion failure by separating the
         *      contiguous code sections into different blocks via two conversion
         *      failure indications for what actually is just one code snippet.
         *      
         *      So if for example, we had 2 lines of code, a separating blank line, then 2
         *      more lines of code. The hash indices for this example would be 0,1,3,4 for
         *      the not retrieved indices from the Ruby expression list that was returned
         *      from the ruby get expression process. The zero hash index list would contain 2,
         *      which is for the blank line. Peace out man/woman/it/thing/whatevurrrrrrr!!!
         *      
         * _retBlkList - 'Retrieved Block List':
         *      A block is set for each Ruby expression corresponding to a retrievable 
         *      mnemonic. Javascript is retrieved as well as the boxes, which are renumbered
         *      per the Ruby code, truncation and line indentation are also restored if present
         *      in the Ruby code. We also add any variables to the class member _varList.
         *      
         * _zeroHashBlkList - A block is set for blank and comment line Ruby expressions. 
         *      These have a hash value of "0". Contiguous expressions [indices] are blocked
         *      together in one block.
        */
        public bool SetBlocks()
        {
            if (CodeRetrieverDataflow.Form1.debug)
            {
                return false;
            }

            /*
             * Caller has ensured that _rubyExprList and _rubyHashList are not null, and 
             * contain at least 1 expression, and also that _rubyExprList.Count is equal to
             * _rubyHashList.Count.
            */
            Block block;
            Block retBlock;
            Block notRetBlock;

            string hashStr;
            string concatHashStr;
            bool processHashes;
            bool foundHash = false;
            int startIndex;
            int stopIndex;
            int lastIndex;
            List<int> nonZeroIndices = new List<int>();
            List<int> notRetIndices = new List<int>();
            Regex commentRegex = new Regex("#");

            try
            {
                _notRetBlkList.Clear();
                _retBlkList.Clear();
                _zeroBlkList.Clear();

                /*
                 * We will first set the list of non-zero hash indices of _rubyHashList.
                 * These are the indices we will use in attempting to retrieve JavaScript for
                 * the corresponding Ruby expressions in _rubyExprList. We also set the
                 * zero hash block list [Ruby comment lines and blank lines] while we are here.
                */
                block = null;

                for (int i = 0; i < _rubyHashList.Count; i++)
                {
                    if (_rubyHashList[i] == "0")
                    {
                        if (block == null)
                        {
                            block = new Block(blockType.zeroHash);
                        }

                        block.hashIndices.Add(i);

                        /*
                         * Expression is either a comment line in which case we convert the
                         * leading '#' to '//' or it is a blank line.
                        */
                        block.outputStr += commentRegex.Replace(_rubyExprList[i], "//", 1);
                    }
                    else
                    {
                        nonZeroIndices.Add(i);

                        if (block != null)
                        {
                            // _rubyHashList indices need to be contiguous for a zero block.
                            _zeroBlkList.Add(block);
                            block = null;
                        }
                    }
                }

                // Handle any hanging zeroBlock:
                if (block != null)
                {
                    // _rubyHashList indices need to be contiguous for a zero block.
                    _zeroBlkList.Add(block);
                }

                /*
                 * We now set _notRetBlkList and _retBlkList by iterating over nonZeroIndices
                 * until all indices are either in one block list or the other.
                */
                startIndex = 0;
                stopIndex = 0;
                lastIndex = nonZeroIndices.Count - 1;

                retBlock = null;
                notRetBlock = null;

                processHashes = (nonZeroIndices.Count > 0);

                while (processHashes)
                {
                    concatHashStr = "";
                    hashStr = "";
                    foundHash = false;

                    for (int i = startIndex; i < nonZeroIndices.Count; i++)
                    {
                        concatHashStr += _rubyHashList[nonZeroIndices[i]];

                        if (i > startIndex)
                        {
                            hashStr = GetMnemonicHash(concatHashStr);
                        }
                        else
                        {
                            hashStr = concatHashStr;
                        }

                        GetJSDataflowByHash(hashStr, ref retBlock);

                        /*
                         * null block means we need to keep checking.
                        */
                        if (retBlock != null)
                        {
                            /*
                             * At this point the block contains the Javascript code for the
                             * associated Ruby expression(s) as well as the generic boxes for
                             * the code snippet. The type has also been set to 'retrieved'.
                            */
                            stopIndex = i;
                            foundHash = true;
                            break;
                        }
                    }

                    if (foundHash)
                    {
                        /*
                         * Check if we have an 'open' notRetBlock:
                        */
                        if (notRetBlock != null)
                        {
                            ProcessNotRetBlock(ref notRetBlock);
                            notRetBlock = null;
                        }

                        /*
                         * If we have a profile instantiation we will put the JS code version into 
                         * _profileList. No need to process this block any further.
                         * 
                         * However, no insertion into _retBlkList causes a small side effect when setting
                         * convertStr under the following situation: If there is a blank line both above
                         * and below a Ruby profile instantiation, 2 blank lines will be present in the
                         * convertStr. We do a 'catch-all' multiple blank line eraser thing in the
                         * method SetConvertStr() which will take care of this.
                         * 
                        */
                        int profileIndex = nonZeroIndices[stopIndex];

                        if (Regex.Match(Regex.Replace(_rubyExprList[profileIndex], @"\s+", ""), @"^(individual=Individual.new|entity=Entity.new|preparer=Preparer.new|fiduciary=Fiduciary.new|equifax=Equifax.new)").Success)
                        {
                            if (!_profileList.Contains(retBlock.javaCode[0] + Environment.NewLine))
                            {
                                _profileList.Add(retBlock.javaCode[0] + Environment.NewLine);
                            }
                        }
                        else
                        {
                            /*
                             * Note that some of these non zero hash indices may not be contiguous.
                             * Two possible reasons: 1) Some of the code snippets actually have a
                             * blank line between code sections, that is the way they were written.
                             * 2) Someone may have inserted some blank lines and/or comment lines in
                             * between code sections.
                            */
                            for (int j = startIndex; j <= stopIndex; j++)
                            {
                                retBlock.hashIndices.Add(nonZeroIndices[j]);
                            }

                            /*
                             * We now need to: Set the generic boxes to their corresponding Ruby
                             * equivalents; Restore any truncation; Add any line indentation that
                             * is present in the Ruby code lines; Update a unique class variable name
                             * list; Insert this block into _retBlkList.
                            */
                            ProcessRubyRetBlock(ref retBlock);

                            if (retBlock == null)
                            {
                                // User has seen the exception message box so we just return.
                                return false;
                            }
                        }

                        startIndex = (stopIndex + 1);
                    }
                    else
                    {
                        /*
                         * We now know there is no correspnding Javascript code for the associated
                         * Ruby expression _rubyExprList[startIndex].
                        */
                        if (notRetBlock == null)
                        {
                            notRetBlock = new Block(blockType.notRetrieved);
                        }

                        notRetBlock.hashIndices.Add(nonZeroIndices[startIndex]);
                        startIndex += 1;
                    }

                    processHashes = (startIndex < nonZeroIndices.Count);
                }

                // Handle any hanging notRetBlock:
                if (notRetBlock != null)
                {
                    ProcessNotRetBlock(ref notRetBlock);
                }

                return true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return false;
            }
        }

        public void ProcessRubyRetBlock(ref Block retBlock)
        {
            string varName = "";
            string leadingWS = "";
            string rubyStr = "";
            string str = "";
            string indivDepNum = null;
            string equSuccOfficNum = null;
            string equPredOfficNum = null;
            string equStateAbbr = null;
            Match wsMatch;
            Match lbMatch;
            Match varMatch;
            MatchCollection matches;

            // retBlock is not null since we are here. So was Kilroy.

            try
            {
                /*
                 * Setup the parts needed to build the output string:
                */

                /*
                 * We need to determine if the associated Ruby code for this block has
                 * any leading white space, if it does we want to keep the white space
                 * indentation. There is no guarantee that the number of lines of Ruby
                 * code is the same as the retrieved JavaScript, not to mention actual
                 * layout in terms of lines of code, etc..so the best we can do
                 * here is to use the first line of Ruby code as our leading white space
                 * indicator.
                */
                wsMatch = Regex.Match(_rubyExprList[retBlock.hashIndices[0]], @"^(\s+)");

                if (wsMatch.Success)
                {
                    leadingWS = wsMatch.Groups[1].Value;
                }

                foreach (int index in retBlock.hashIndices)
                {
                    rubyStr += _rubyExprList[index];
                }

                //MessageBox.Show(string.Format("ProcessRubyRetBlock() - rubyStr:\n\n{0}", rubyStr));

                // Looking for '#' type values:
                lbMatch = Regex.Match(rubyStr, @".dependent(\d)_");

                if (lbMatch.Success)
                {
                    indivDepNum = lbMatch.Groups[1].Value;
                }
                else
                {
                    lbMatch = Regex.Match(rubyStr, @".succ_officer(\d)_");

                    if (lbMatch.Success)
                    {
                        equSuccOfficNum = lbMatch.Groups[1].Value;
                    }
                    else
                    {
                        lbMatch = Regex.Match(rubyStr, @".pred_officer(\d)_");

                        if (lbMatch.Success)
                        {
                            equPredOfficNum = lbMatch.Groups[1].Value;
                        }
                        else
                        {
                            lbMatch = Regex.Match(rubyStr, @"equifax.([A-Z]{2})_");

                            if (lbMatch.Success)
                            {
                                equStateAbbr = lbMatch.Groups[1].Value;
                            }
                        }
                    }
                }

                /*
                 * Set the list of Ruby boxes, 'as discovered' and avoid duplicates.
                */
                matches = Regex.Matches(rubyStr, @"box\d+_\d+");

                // Iterate over the collection of box matches:
                foreach (Match match in matches)
                {
                    // MessageBox.Show(String.Format("match.Value: {0}", match.Value));
                    if (!retBlock.rubyBoxes.Contains(match.Value))
                    {
                        retBlock.rubyBoxes.Add(match.Value);
                    }
                }

                /*
                * Now build the output string:
                */

                /*
                 * Set the block's output string reflecting Ruby's box numbers, truncation,
                 * and any leading white space chars. We also reset any '#' proxy chars to the
                 * actual values from the Ruby code, see up above.
                 * 
                 * Note that we need to skip over any comment lines that may precede the actual
                 * JS code as we want to retain the original comments, and we need to restore the
                 * newline char.
                 * 
                 * We will also update the class _varList with any new JS variable declarations
                 * if required.
                */
                foreach (string codeLine in retBlock.javaCode)
                {
                    if (!Regex.Match(codeLine, @"^//").Success)
                    {
                        str += leadingWS + codeLine + Environment.NewLine;

                        if (_varFlag)
                        {
                            /*
                             * The (?i) says to ignore case. Did this because I was stubbing in Ruby
                             * code that was upper cased to simulate the JS code. Leaving this in since
                             * some vars in the JS use upper cased chars. [camel casing I guess].
                             */
                            varMatch = Regex.Match(codeLine, @"(?i)([a-z][a-z0-9_]+)\s+=\s+.+");

                            // Original code for var indentification.
                            //varMatch = Regex.Match(codeLine, @"(?i)^\s*([a-z][a-z0-9_]+)\s*=\s*.+");

                            if (varMatch.Success)
                            {
                                varName = string.Format("var {0} = '';{1}", varMatch.Groups[1].Value, Environment.NewLine);

                                if (!_varList.Contains(varName))
                                {
                                    _varList.Add(varName);
                                }
                            }
                        }
                    }
                }

                for (int i = 0; i < retBlock.javaBoxes.Count; i++)
                {
                    str = str.Replace(retBlock.javaBoxes[i], retBlock.rubyBoxes[i].ToUpper());
                }

                str = str.Replace("BOX", "box");

                /*
                 * Now reset those '#' chars to their real values:
                */
                if (indivDepNum != null)
                {
                    str = Regex.Replace(str, @".dependent#_", string.Format(".dependent{0}_", indivDepNum));
                }
                else if (equSuccOfficNum != null)
                {
                    str = Regex.Replace(str, @".succ_officer#_", string.Format(".succ_officer{0}_", equSuccOfficNum));
                }
                else if (equPredOfficNum != null)
                {
                    str = Regex.Replace(str, @".pred_officer#_", string.Format(".pred_officer{0}_", equPredOfficNum));
                }
                else if (equStateAbbr != null)
                {
                    str = Regex.Replace(str, @"equifax.##_", string.Format("equifax.{0}_", equStateAbbr));
                }

                // Add truncation:
                if (retBlock.truncChk)
                {
                    // Clipboard.SetText first clears the Clipboard.
                    Clipboard.SetText(rubyStr, TextDataFormat.Text);

                    _truncList = _rubyExp.GetTruncationData();

                    if ((_truncList != null) && (_truncList.Count > 0))
                    {
                        string[] parts;
                        string errStr = "";
                        string printLine = "";

                        /*
                         * The truncation logic here is not the same as for the retrieve JS code,
                         * retrieve JS is way looser on the truncation specified by the user whereas
                         * the conversion code checks an XML attribute to see if truncation is even allowed
                         * for a snippet.
                         */
                        foreach (string boxInfo in _truncList)
                        {
                            parts = boxInfo.Split('|');

                            //MessageBox.Show(String.Format("parts[0]: {0}\nparts[1]: {1}", parts[0], parts[1]));

                            if (parts[1] == "DIFF")
                            {
                                errStr += string.Format("// CONVERSION TRUNCATION WARNING: TRUNCATION FOR {0} CANNOT BE DETERMINED.{1}", parts[0], Environment.NewLine);
                            }
                            else
                            {
                                printLine = string.Format(@"\s*SFForms.StorageManager.setValue\(formInstanceId\s*,\s*[""']{0}[""']\s*,\s*[a-zA-Z0-9_\[\]']+", parts[0]);

                                str = Regex.Replace(str, printLine, delegate(Match match)
                                {
                                    return match.ToString() + string.Format(".substr(0,{0})", parts[1]);
                                });
                            }
                        }

                        // Alert user if the Ruby script had trouble identifying some truncation:
                        if (errStr.Length > 0)
                        {
                            str = errStr + str;
                        }
                    }
                }

                retBlock.outputStr = str;

                _retBlkList.Add(retBlock);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                retBlock = null;
            }
        }

        /*
         * All we need do here is to wrap the Ruby expression(s) in a comment
         * and insert the block into _notRetBlkList.
        */
        //public void ProcessNotRetBlock(ref Block notRetBlock)
        //{
        //    notRetBlock.outputStr = "/* CONVERSION FAILURE" + Environment.NewLine;

        //    foreach (int index in notRetBlock.hashIndices)
        //    {
        //        notRetBlock.outputStr += _rubyExprList[index];
        //    }

        //    notRetBlock.outputStr += "*/" + Environment.NewLine;

        //    _notRetBlkList.Add(notRetBlock);
        //}

        /*
         * Trying a non-contiguous notRetBlock approach. We INCLUDE zero hash Ruby
         * expressions as long as they lie between the first index and the last index.
         * We also REMOVE these zero hash expressions from the associated _zeroBlkList
         * block, and will remove the block if we empty it's hash index list.
        */
        public void ProcessNotRetBlock(ref Block notRetBlock)
        {
            notRetBlock.outputStr = "/* CONVERSION FAILURE" + Environment.NewLine;

            for (int i = notRetBlock.hashIndices[0]; i <= notRetBlock.hashIndices[notRetBlock.hashIndices.Count - 1]; i++)
            {
                notRetBlock.outputStr += _rubyExprList[i];

                if (!notRetBlock.hashIndices.Contains(i))
                {
                    // Remove this index value from the appropriate _zeroBlkList block.
                    for (int j = 0; j < _zeroBlkList.Count; j++)
                    {
                        if (_zeroBlkList[j].hashIndices.Contains(i))
                        {
                            /*
                             * The way we structure this list means we do not need to worry about breaking
                             * the contiguous indices setup.
                            */
                            _zeroBlkList[j].hashIndices.Remove(i);

                            if (_zeroBlkList[j].hashIndices.Count == 0)
                            {
                                _zeroBlkList.RemoveAt(j);
                                break;
                            }
                        }
                    }
                }
            }

            //notRetBlock.outputStr += "*/" + Environment.NewLine;
            notRetBlock.outputStr += (notRetBlock.outputStr.EndsWith(Environment.NewLine) ? "" : Environment.NewLine) + "*/" + Environment.NewLine;
            _notRetBlkList.Add(notRetBlock);
        }

        public void GetJSDataflowByHash(string hashStr, ref Block block)
        {
            bool notFound = true;

            if (xmlDoc != null)
            {
                IEnumerable<XElement> nodes;
                IEnumerable<XElement> subNodes;
                XElement node = null;

                try
                {
                    /*
                     * Using .Contains because there may be multiple hashes for the same JavaCodeSnippet.
                     * This would mean we ran into older code styles that are not used anymore, but we
                     * will still convert them to the JavaScriptCode version.
                    */
                    nodes = xmlDoc.Descendants().Where(e => ((e.Name == "Block") && e.HasAttributes && (e.Attribute("hashId") != null) && (e.Attribute("hashId").Value.Contains(hashStr))));

                    if (nodes.Count() > 0)
                    {
                        notFound = false;
                        block = new Block(blockType.retrieved);
                        node = nodes.First();

                        subNodes = node.Elements().Where(e => ((e.Name == "Boxes") || (e.Name == "JavaBoxes") || (e.Name == "JavaScriptCode")));

                        foreach (XElement e in subNodes)
                        {
                            if (e.Name == "Boxes")
                            {
                                block.javaBoxes = e.Value.Split(',').ToList();
                            }
                            /*
                             * I think we will always have "Boxes" before "JavaBoxes" from the xml structure.
                             * This is the rare case where we need to change the box order else the renumbering
                             * to Ruby boxes would be wrong.
                            */
                            else if (e.Name == "JavaBoxes")
                            {
                                // Reassign to JavaScript's box ordering - rare.
                                block.javaBoxes = e.Value.Split(',').ToList();
                            }
                            else if (e.Name == "JavaScriptCode")
                            {
                                block.javaCode = e.Value.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None).ToList();

                                /*
                                 * Do we bother to check for truncation? Doing it this way because
                                 * some older Ruby code styles often used sub-string coding to
                                 * dataflow to validated fields like SSNs, Phone Numbers, etc.,
                                 * which 'looks like' our truncation, so this helps clarify the
                                 * situation logic wise.
                                */
                                if (e.HasAttributes && (e.Attribute("truncChk") != null) && (e.Attribute("truncChk").Value == "Y"))
                                {
                                    block.truncChk = true;
                                }
                            }
                            else
                            {
                                // Should not happen once we build the XML file.
                                MessageBox.Show(string.Format("Unknown node detected for hash:\n\n{0}\n\nNode Name: {1}", hashStr, e.Name), "Unknown Node Detected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                    MessageBox.Show(e.ToString());
                }
            }
            else
            {
                //MessageBox.Show("xmlBuffer is empty.", "No XML Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
                MessageBox.Show("xmlDoc is null.", "GetJSDataflowByHash(): No XML Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            if (notFound)
            {
                block = null;
            }
        }

        /*
         * Calculates the actual (mnemonic) hash value for a Retriever code snippet. We had to
         * have first converted the actual Ruby code representing the snippet via a call to
         * DoConversion() which set the _rubyHashList. Each hash in this list corresponds
         * to a complete Ruby code expression. The caller then concatenated all non-zero
         * hashes in _rubyHashList together which we use here to calculate the final hash which 
         * represents the code snippet.
        */
        public string GetMnemonicHash(string concatHash)
        {
            string hashStr = null;

            if (!string.IsNullOrEmpty(concatHash))
            {
                hashStr = md5(Encoding.ASCII.GetBytes(concatHash)).ToString();
            }

            return hashStr;
        }

        public List<string> GetExprList()
        {
            return _rubyExprList;
        }

        public List<string> GetHashList()
        {
            return _rubyHashList;
        }

        public List<string> GetFilterList()
        {
            return _filterList;
        }

        #region PrivateMethods

        private void InitLists()
        {
            _varList.Clear();
            _profileList.Clear();
            _truncList.Clear();
            _rubyExprList.Clear();
            _rubyHashList.Clear();
            _filterList.Clear();
            _notRetBlkList.Clear();
            _retBlkList.Clear();
            _zeroBlkList.Clear();
        }

        private ulong md5(byte[] Source)
        {
            if (Source.Length == 0)
                return 0;

            HashAlgorithm ha = GetHashAlgorithm("MD5");

            byte[] computed_hash = ha.ComputeHash(Source);

            ulong ret = BitConverter.ToUInt64(computed_hash, 0);

            return ret;
        }

        private HashAlgorithm GetHashAlgorithm(string HashAlgorithmName)
        {
            switch (HashAlgorithmName.ToUpperInvariant())
            {
                case "MD5":
                    return MD5.Create();
                case "SHA1":
                    return SHA1.Create();
                case "SHA256":
                    return SHA256.Create();
                default:
                    throw new ArgumentOutOfRangeException("HashType", "Don't know how to create a '" + HashAlgorithmName + "' hash.");
            }
        }

        private bool ComputeExprHashes()
        {
            // Caller has ensured _rubyExprList is not null or empty.
            ulong md5Hash;
            string filteredExpr;
            bool ret = true;

            try
            {
                foreach (string expression in _rubyExprList)
                {
                    // Overboard??
                    if (expression != null)
                    {
                        filteredExpr = expression;

                        FilterRubyCode(ref filteredExpr);

                        md5Hash = md5(Encoding.ASCII.GetBytes(filteredExpr));

                        _rubyHashList.Add(md5Hash.ToString());

                        // Debug Only:
                        _filterList.Add(filteredExpr);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                ret = false;
            }

            return ret;
        }

        private void FilterRubyCode(ref string rubyCode)
        {
            // Caller has ensured rubyCode is not null.

            string filteredStr = "";
            string tmpStr = "";
            string[] linesArray = rubyCode.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

            //MessageBox.Show(String.Format("linesArray.Length = {0}", linesArray.Length));

            /*
             * Using lastChar to check if these chars have been escaped, if so they lose
             * their special meaning in the below foreach loop.
            */
            bool singleQuote = false;
            bool doubleQuote = false;
            bool forSlash = false;

            char lastChar = ' ';

            //int lineCnt = 1;
            foreach (string line in linesArray)
            {
                //MessageBox.Show(String.Format("line {0}:{1}", lineCnt, line));

                // Ignore comment lines.
                if (!Regex.Match(line, @"^\s*#").Success)
                {
                    tmpStr += line;
                }

                //lineCnt += 1;
            }

            if (tmpStr.Length > 0)
            {
                // Remove tabs and newlines.
                tmpStr = Regex.Replace(tmpStr, @"[\t\n]+", "");

                /*
                 * Remove substring code: for example, remove [0,10]. Also remove
                 * substrings like [-2, 2], [  -  2 , 2 ], etc. Ruby does not allow 
                 * the second value to be negative.
                */
                tmpStr = Regex.Replace(tmpStr, @"\[\s*-?\s*\d+\s*,\s*\d+\s*\]", "");

                // Rename all boxes to box: for example 'box10_15' becomes 'box'.
                tmpStr = Regex.Replace(tmpStr, @"box\d+_\d+", "box");

                /*
                 * Reset any individual dependent code to the generic '#' placeholder for the
                 * numbers 1, 2, 3, 4, 5, or 6. For example,
                 * 'tmp = individual.dependent2_first_name.strip' is changed to the snippet
                 * file version 'tmp = individual.dependent#_first_name.strip'.
                */
                tmpStr = Regex.Replace(tmpStr, @".dependent\d_", ".dependent#_");

                /*
                 * Reset any equifax successor and predecessor officer code to the
                 * generic '#' placeholder for the numbers 1, 2, or 3.
                 * For example, 'tmp = equifax.succ_officer3_dob.strip' is changed to the
                 * snippet file version 'tmp = equifax.succ_officer#_dob.strip'.
                */
                tmpStr = Regex.Replace(tmpStr, @".succ_officer\d_", ".succ_officer#_");
                tmpStr = Regex.Replace(tmpStr, @".pred_officer\d_", ".pred_officer#_");

                /*
                 * Reset any equifax state list code to the generic '##' placeholder
                 * for the uppercase 2 letter state abbreviations. We will just look for 2
                 * uppercase letters. For example,
                 * 'tmp = equifax.NY_pred_unemp_acct_number.strip' is changed to the snippet
                 * file version 'tmp = equifax.##_pred_unemp_acct_number.strip'.
                */
                tmpStr = Regex.Replace(tmpStr, @"equifax.[A-Z]{2}_", "equifax.##_");

                /*
                 * Remove space chars [' '] where appropriate. i.e. Not in a regular 
                 * expression, a delete(), etc..
                */
                foreach (char c in tmpStr)
                {
                    switch (c)
                    {
                        case ('\''):
                            {
                                //MessageBox.Show("Got a single quote.");
                                if ((!doubleQuote) && (!forSlash) && (lastChar != '\\'))
                                {
                                    singleQuote = !singleQuote;
                                }
                                break;
                            }

                        case ('\"'):
                            {
                                //MessageBox.Show("Got a double quote.");
                                if ((!singleQuote) && (!forSlash) && (lastChar != '\\'))
                                {
                                    doubleQuote = !doubleQuote;
                                }
                                break;
                            }

                        case ('/'):
                            {
                                //MessageBox.Show("Got a forward slash.");
                                if ((!singleQuote) && (!doubleQuote) && (lastChar != '\\'))
                                {
                                    forSlash = !forSlash;
                                }
                                break;
                            }

                        case (' '):
                            {
                                //MessageBox.Show("Got a space.");
                                if (!(singleQuote || doubleQuote || forSlash))
                                {
                                    continue;
                                }
                            }
                            break;
                    }

                    lastChar = c;
                    filteredStr += c;
                }
            }
            else
            {
                filteredStr = tmpStr;
            }

            rubyCode = filteredStr;
        }

        #endregion

        private List<string> _varList;
        private List<string> _profileList;
        private List<string> _truncList;
        private List<string> _rubyExprList;
        private List<string> _rubyHashList;
        private List<Block> _notRetBlkList;
        private List<Block> _retBlkList;
        private List<Block> _zeroBlkList;
        private RubyExpression _rubyExp;
        private bool _varFlag;

        public byte[] xmlBuffer;
        public XDocument xmlDoc;
        public string convertStr;
        public bool conversionError;

        // Debug Only:
        private List<string> _filterList;
    }

    public class Block
    {
        public Block(blockType type)
        {
            this.type = type;
            outputStr = "";
            hashIndices = new List<int>();
            javaBoxes = new List<string>();
            rubyBoxes = new List<string>();
            javaCode = new List<string>();
            truncChk = false;
        }

        // Not sure if I need all these getters and Irish setters.
        public blockType type { get; set; }
        public string outputStr { get; set; }
        public List<int> hashIndices { get; set; }
        public List<string> javaBoxes { get; set; }
        public List<string> rubyBoxes { get; set; }
        public List<string> javaCode { get; set; }
        public bool truncChk { get; set; }
    }
}